#!/usr/bin/env python
# coding: utf-8

import numpy as np

class Statistica:
   
    # Наблюдаемое значение критерия Пирсона для проверки гипотезы о равномерности распределения
    # f_observed - матрица наблюдаемых частот попадания в подобласти
    @staticmethod
    def pirsonCriterion(f_observed):        
        # общее число наблюдений:
        n = sum(sum(f_observed))
        # число интервалов:
        (nx,ny) = np.shape(f_observed)
        k = nx*ny
        # теоретическая частота:
        f_teor = n/k
        # Наблюдаемое значение критерия Пирсона:
        pirs_obs = sum(sum((f_observed - f_teor)**2/f_teor))
        
        return pirs_obs        
    
    @staticmethod
    def uniformity(blobs, height, width, sizeBlock):
        heightBlocks = math.ceil(height/sizeBlock)
        widthBlocks = math.ceil(width/sizeBlock)     
        
        if ( (sizeBlock*heightBlocks != height) or (sizeBlock*widthBlocks != width) ):
            print("\nWARNING: Высота или длина изображения не делится на заданный размер блока нацело!\n")
        
        counter = np.zeros((heightBlocks, widthBlocks), dtype=int)
        
        for blob in blobs:
            y,x,r = blob
            i = math.ceil(y/sizeBlock)-1
            j = math.ceil(x/sizeBlock)-1
            counter[i, j] += 1
        
        print("Размер подобласти (NxN) в пикселях:", sizeBlock, "\tЧисло подобластей:", heightBlocks*widthBlocks)
        pirs_obs = Statistica.pirsonCriterion(counter)
        print("Наблюдаемое значение критерия Пирсона:", pirs_obs)
        print("Наблюдаемое значение критерия Пирсона, деленное на число подобластей:",
              pirs_obs/(heightBlocks*widthBlocks))       
        
        return counter

    @staticmethod
    def printStat(arr):
        print('\tчисло различных значений:', len(set(arr)))
        print('\tминимум:', np.min(arr))
        print('\tмаксимум:', np.max(arr))      
        print('\tсреднее:', np.mean(arr))              
        print('\tмедиана:', np.median(arr))
        print('\tСКО:', np.std(arr))
    
    @staticmethod
    def showRadiusInformation(blobs):
        R = blobs[:, 2]
        print('Всего найдено наночастиц: ', blobs.shape[0])
        print('Cтатистика по размерам:')
        Statistica.printStat(R)
        radius, count = np.unique(R, return_counts=True)
        print('Количество частиц каждого из радиусов (радиус:количество частиц)')
        print([str(np.round(radius[i], 2))+":"+str(count[i]) for i in range(0, len(radius))])
    
    # Евклидово расстояние
    @staticmethod
    def findEuclideanDistances(blobs):
        nblobs = np.shape(blobs)[0]
        dist = np.zeros((nblobs, nblobs),dtype = float)
        for i in range(nblobs):
            for j in range(nblobs):
                dist[i,j] = np.sqrt((blobs[i,0]-blobs[j,0])**2 + (blobs[i,1]-blobs[j,1])**2)
        return dist
    
    # Расстояние Чебышева
    @staticmethod
    def findModulDistances(blobs):
        nblobs = np.shape(blobs)[0]
        dist = np.zeros((nblobs, nblobs),dtype = float)
        for i in range(nblobs):
            for j in range(nblobs):
                dist[i,j] = max(np.abs(blobs[i,0]-blobs[j,0]), np.abs(blobs[i,1]-blobs[j,1]))
        return dist
    
    # 
    # distance - набор расстояний, для которых считается доп. статистика
    #    о числе ближайших частиц (если список пуст, статистика не считается)
    @staticmethod
    def distance(blobs):
        dist = Statistica.findEuclideanDistances(blobs)
        
        nblobs = np.shape(blobs)[0]
        minDist = np.min(dist+np.eye(nblobs, nblobs)*99999,0)
        
        print('Статистика по расстояниям до ближайшей наночастицы:')
        Statistica.printStat(minDist)        
        
        return dist, minDist


from Accuracy import Accuracy
from blobs2roi import blobs2roi
import csv
import numpy

def jakkarAccuracy(pathFileGT, pathFileEst, roi=None, threshold=0.25):

    with open(pathFileGT, 'r') as FileGT:
        reader = csv.reader(FileGT)
        gt_blobs = numpy.array(list(reader), dtype=float)

    with open(pathFileEst, 'r') as FileEst:
        reader = csv.reader(FileEst)
        est_blobs = numpy.array(list(reader), dtype=float)

    if (roi==None):
        roi = blobs2roi(gt_blobs, 960, 1280)
        print("Warning! TODO: размеры изображения!")
    
    return Accuracy.Acc(gt_blobs, est_blobs, roi, threshold)


#!/usr/bin/env python
# coding: utf-8

import numpy as np

class Accuracy:
    @staticmethod
    def blobs_in_roi(blobs, roi):
        """Check if the center of blob is inside ROI  

        Arguments
        blobs -- list or array of areas oссupied by the nanoparticle 
                (y, x, r) y and x are coordinates of the center and r - radius    
        roi -- (y,x,h,w)

        Return blobs list
        """
        indexes = list(map(lambda blob: int(blob[0]) >= roi[0]                                     and int(blob[1]) >= roi[1]                                     and int(blob[0]) < roi[0]+roi[2]                                      and int(blob[1]) < roi[1]+roi[3],                                         blobs))
        return np.copy(blobs[indexes])

    @staticmethod
    def findIOU4circle(c1, c2):
        """Finds Jaccard similarity measure for two circles, 
           defined by the coordinates of centers and radii.
           c1=[x1,y1,r1], c2=[x2,y2,r2]  
        """

        d = np.linalg.norm(c1[:2] - c2[:2]) #distance betweem centers

        rad1sqr = c1[2] ** 2
        rad2sqr = c2[2] ** 2

        if d == 0:
            # the circle centers are the same
            return min(rad1sqr, rad2sqr)/max(rad1sqr, rad2sqr)

        angle1 = (rad1sqr + d ** 2 - rad2sqr) / (2 * c1[2] * d)
        angle2 = (rad2sqr + d ** 2 - rad1sqr) / (2 * c2[2] * d)

        # check if the circles are overlapping
        if (-1 <= angle1 < 1) or (-1 <= angle2 < 1):
            theta1 = np.arccos(angle1) * 2
            theta2 = np.arccos(angle2) * 2

            area1 = (0.5 * theta2 * rad2sqr) - (0.5 * rad2sqr * np.sin(theta2))
            area2 = (0.5 * theta1 * rad1sqr) - (0.5 * rad1sqr * np.sin(theta1))

            return (area1 + area2)/(np.pi*(rad1sqr+rad2sqr) - area1 - area2)

        elif angle1 < -1 or angle2 < -1:
            # Smaller circle is completely inside the largest circle.
            # Intersection area will be area of smaller circle
            # return area(c1_r), area(c2_r)
            return min(rad1sqr, rad2sqr)/max(rad1sqr, rad2sqr)
        return 0

    @staticmethod
    def СonformBlobs(blobs_gt, blobs_est, roi, thres):
        blobs_gt = Accuracy.blobs_in_roi(blobs_gt, roi)
        blobs_est = Accuracy.blobs_in_roi(blobs_est, roi)  

        length_gt = blobs_gt.shape[0]
        length_est = blobs_est.shape[0]

        iou = np.zeros((length_gt, length_est))
        for i in range(length_gt):
            for j in range(length_est):
                iou[i,j] = Accuracy.findIOU4circle(blobs_gt[i], blobs_est[j])

        diff_blobs = []

        for i in range(length_gt):
            if (max(iou[i]) >= thres):
                imax = np.argmax(iou[i])
                if (i == np.argmax(iou[:, imax])):            
                    diff_blobs.append(blobs_gt[i] - blobs_est[imax])

        return diff_blobs
    
    @staticmethod
    def AccInfo(blobs_gt, blobs_est, roi, thres):
        blobs_gt = Accuracy.blobs_in_roi(blobs_gt, roi)
        blobs_est = Accuracy.blobs_in_roi(blobs_est, roi)  

        length_gt = blobs_gt.shape[0]
        length_est = blobs_est.shape[0]

        iou = np.zeros((length_gt, length_est))
        for i in range(length_gt):
            for j in range(length_est):
                iou[i,j] = Accuracy.findIOU4circle(blobs_gt[i], blobs_est[j])

        match = 0
        no_match = 0
        fake = 0
        no_match_index = np.zeros(length_gt,dtype = 'bool')

        match_matr = np.zeros((length_gt, length_est), dtype = int)

        for i in range(length_gt):
            if max(iou[i])>=thres:
                imax = np.argmax(iou[i])
                match_matr[i,imax] = 1

        fake_index = np.zeros(length_est,dtype = 'bool')
        for j in range(length_est):
            if sum(match_matr[:,j])>1: 
                imax = np.argmax(iou[:,j])
                match_matr[:, j] = np.zeros(length_gt, dtype = int)
                match_matr[imax, j] = 1 
            if sum(match_matr[:,j]) == 0:
                fake+=1
                fake_index[j] = True
        fake_blobs = blobs_est[fake_index]

        for i in range(length_gt): 
            if sum(match_matr[i,:]) == 0: 
                no_match_index[i] = True

        no_match = sum(no_match_index)
        match = sum(sum(match_matr))        
        no_match_gt_blobs =  blobs_gt[no_match_index]       

        return match, no_match, fake, no_match_gt_blobs, fake_blobs
    
    @staticmethod
    def Acc(blobs_gt, blobs_est, roi, thres):
        blobs_gt = Accuracy.blobs_in_roi(blobs_gt, roi)
        blobs_est = Accuracy.blobs_in_roi(blobs_est, roi)  

        length_gt = blobs_gt.shape[0]
        length_est = blobs_est.shape[0]

        iou = np.zeros((length_gt, length_est))
        for i in range(length_gt):
            for j in range(length_est):
                iou[i,j] = Accuracy.findIOU4circle(blobs_gt[i], blobs_est[j])

        match = 0
        for i in range(length_gt):
            if (max(iou[i]) >= thres):
                imax = np.argmax(iou[i])
                if (i == np.argmax(iou[:, imax])):            
                    match = match + 1 
                            
        return match/(length_gt+length_est-match)
    
    @staticmethod
    def TableMatchAlgoritms(reviewAreas:list, roi):
        countAreas = len(reviewAreas)
        table = np.ones([countAreas, countAreas],'float')

        for i in range(countAreas):
            for j in range(i, countAreas):
                if (i != j):                                   
                    table[i,j] = Accuracy.Acc(reviewAreas[i], reviewAreas[j], roi, 0.25) 
                    table[j,i] = table[i,j]

        return table


#!/usr/bin/env python
# coding: utf-8

import numpy as np

def blobs2roi(_blobs, _heightImg, _widthImg):
    roi = np.zeros(4, dtype='int')
    roi[0] = max(0, (_blobs[:,0]-_blobs[:,2]).min()) 
    roi[1] = max(0, (_blobs[:,1]-_blobs[:,2]).min())
    roi[2] = min(_heightImg, (_blobs[:,0]+_blobs[:,2]).max() - roi[0]+1)
    roi[3] = min(_widthImg, (_blobs[:,1]+_blobs[:,2]).max() - roi[1]+1)
    return roi 



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="CTSS",
    version="0.1.0",
    author="Mikhail Kurbakov",
    author_email="muwa1997@mail.ru",
    description="Software package for the CT&SS laboratory of Tula State University",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=['numpy>=1.21'],
    url="https://github.com/muwsik",
    #project_urls={
    #    "tag": "url",
    #},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8.10",
)


#!/usr/bin/env python
# coding: utf-8

import pandas as pd
import numpy as np

# Приведение разметки экспертов к унифицированному виду: "центр_по_У, центр_по_Х, радиус"
def gt2blobs(_pathCoordGT, _pathRadiiGT, _heightImg):
    gt_coords = pd.read_excel(_pathCoordGT, header = None, nrows = 2, engine='openpyxl')
    gt_unit = pd.read_csv(_pathRadiiGT, sep = ';')

    # replace не работае, если вещественное число уже записано через точку
    temp_length_pix = gt_unit[gt_unit['Unit']=='pixels']['Length'][0]
    if (not isinstance(temp_length_pix, float)):    
        length_pix = float(temp_unit.replace(',', '.'))
    else:
        length_pix = temp_length_pix
    length_nm = np.linalg.norm(gt_coords[[0,1]].diff(),axis=1)[1]
    coeff_nm2pix = length_pix/length_nm
    
    gt_blobs = []
    for col in range(2,gt_coords.shape[1],2):
        line = gt_coords[[col, col+1]]
        x,y = np.array(line.mean())*coeff_nm2pix
        r = np.linalg.norm(line.diff(),axis=1)[1]*coeff_nm2pix / 2.
        gt_blobs.append([_heightImg-y,x,r])
    gt_blobs = np.array(gt_blobs)
    
    return gt_blobs



#!/usr/bin/env python
# coding: utf-8

import matplotlib.pyplot as plt

class Graphics:

    @staticmethod
    def createTwoPlot(image, blobs, title = "", flagPaintParticles = False, color = 'blue', size = 6):
        fig, (ax_im, ax_second) = plt.subplots(1, 2, figsize=(size*2,size), sharex=False, sharey=False)

        ax_im.imshow(image, cmap='gray', aspect='auto')       
        ax_im.set_axis_off()        
        ax_im.set_title(title)

        if (flagPaintParticles):
            for blob in blobs:
                y, x, r = blob
                c = plt.Circle((x, y), r, color=color, linewidth=1, fill=False)
                ax_im.add_patch(c)

        plt.tight_layout()
        return ax_second
    
    # Формирует столбчатую диаграмму
    # array - полный набор данных, по которому строится диаграмма
    # axis - оси в которых будет отображаться диаграмма
    # gap - процентное отношение ширины столбца к ширине зазора между соседними столбцами
    # is_norm - нормированная диаграмма (по умолчанию не нормированная)
    @staticmethod
    def createBar(array, axis, gap=0.08, is_norm=False):
        value, hieght = np.unique(array, return_counts=True)
        width = np.diff(value)
        width = np.append(width, width[-1])   

        if (is_norm):
            axis.bar(value, hieght/len(array), width - width*gap, alpha = 0.5)
            axis.set_ylim(0, 1)
        else:
            axis.bar(value, hieght, width - width*gap, alpha = 0.5)

        axis.set_xticks(value)
        axis.set_xticklabels(["{0:.0f}".format(value[i]) for i in range(0, len(value))])
